/////////////////////////////////////////////////////////////////////////////
// dumphex v0.5
// ----------
// (c) 1999..2007 by Dirk Jagdmann <doj@cubic.org>
//
// this program prints files/stdin in a hexedit form to stdout. Usefull when
// you want to have a look at a file.
//
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>

// the version string for the current source code
static const char *VERSION="v0.5";

/////////////////////////////////////////////////////////////////////////////
// static int dump(FILE *f, bool beginMode=false, bool endMode=false)
// ----------------------------------------------
// this function dumps the file pointed by FILE *f to stdout in a hexlist
// form. If an error is encountered while reading or positioning the stream
// the function is aborted.
//
// parameters: FILE *f       - the stream to parse
//             bool beginMode - enable parsing the beginning of the stream
//             bool endMode - enable parsing the end of the stream
//
// configuration: two global variables
//                tailLength - when tailmode is true the number of bytes to
//                             read
//                hexWidth   - number of bytes in each row
//
// return value: >=0 the number of bytes successfully read from the stream
//                -1 upon error in the input stream
/////////////////////////////////////////////////////////////////////////////
// the number of bytes to be used when begin/endMode is used
static int dumpLength=256;

// the width of the dumped output
static int bytePerRow=16;

static int dump(FILE *f, bool beginMode=false, bool endMode=false) {
  // the offset in the input stream. Used only to print
  int offset=0;

  // the number of bytes ro read. -1 means oo
  int offsetEnd=beginMode?dumpLength:-1;

  // look if the tailmode is enabled. If so we position the stream near the end
  if(endMode) {
    // seek near the end
    if(fseek(f, -dumpLength, SEEK_END)==-1) {
      fprintf(stderr,"dump: error while seeking\n");
      return -1;
    }
    // get current position
    offset=ftell(f);
    // error check, if position is valid
    if(offset==-1) {
      fprintf(stderr,"dump: error while telling\n");
      return -1;
    }
  }

  // one row on the display
  unsigned char *buf=new unsigned char[bytePerRow];
  if(buf==0) {
    fprintf(stderr, "dump: error while allocating memory\n");
    return -1;
  }

  // read until the stream has finished
  while(offset!=offsetEnd) {
    int i;

    // read one row
    int read=fread(buf, 1, bytePerRow, f);
    // error check for the above fread
    if(ferror(f)) {
      fprintf(stderr,"dump: error while reading\n");
      delete [] buf;
      return -1;
    }
    if(read<=0)
      break;

    // print offset in the file
    if(bytePerRow==16)
      printf("%08X: ", offset);
    else
      printf("%06X: ", offset);

    // print the bytes in hex
    for (i=0; i<read; ++i)
      printf("%02X ",buf[i]);
    while(i++<bytePerRow)
      printf("   ");

    if(bytePerRow==16)
      printf("- ");

    // print the bytes again, as string
    for (i=0; i<read; ++i)
      printf("%c", isprint(buf[i])?buf[i]:'.');

    printf("\r\n");

    offset+=read;
  }

  delete [] buf;
  return offset;
}

/////////////////////////////////////////////////////////////////////////////
// int main(int argc, char **argv)
// --------------------------------
// this function parses the command line for options and then calls the
// routine "static int dump(...)" for each stream given at the command line
//
// return value:  0 upon succes. No stream failed
//                1 upon partial success. Some streams failed
//                2 upon command line error
/////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  // this variable enables/disables the begin/end mode
  bool beginMode=false;
  bool endMode=false;

  // parse the command line
  int opt;
  while((opt=getopt(argc, argv, "beht8w:")) >= 0)
    switch(opt) {
      case 'h': printf( "dumphex %s (c) by Dirk Jagdmann <doj@cubic.org>\n"
                        "a program to print data in hex format\n"
                        "dump: -h  help text\n"
                        "      -b  begin mode. show first %i/%i bytes of input\n"
                        "      -e  end mode. show last %i/%i bytes of input\n"
                        "      -e  tail mode. same as end mode\n"
                        "      -8  output 8 bytes in a row (default is 16 byte per row)\n"
			"      -w <width> output width bytes in a row\n"
                        , VERSION
                        , dumpLength, dumpLength/2
                        , dumpLength, dumpLength/2
                      );
                return 0;
                break;

      case 'b': beginMode=true;
                endMode=false;
                break;

      case 'e':
      case 't': endMode=true;
                beginMode=false;
                break;

      case '8': bytePerRow=8; dumpLength/=2; break;

      case 'w':
	bytePerRow=atoi(optarg);
	if(bytePerRow<=0)
	  {
	    fprintf(stderr, "output width must be positive and not %i\n", bytePerRow);
	    exit(1);
	  }
	break;

      case '?': return 2;
    }

  // return value for the main routine
  int ret=0;

  // when no options are given at the command line stdin is used as stream
  if(optind==argc) {
    if(beginMode || endMode)
      printf("dumping %i bytes of stdin\n", dumpLength);
    else
      printf("dumping stdin\n");
    int read=dump(stdin, beginMode, endMode);
    if(read==-1) {
      fprintf(stderr,"dump: error while dumping stdin\n");
      ret=1;
    }
    else {
      printf("read %i bytes\n\n", read);
    }
  }
  // otherwise we parse the remaining options for filenames
  else {
    for(int argnum=optind; argnum<argc; argnum++) {
      // try to open the file
      FILE *f=fopen(argv[argnum],"rb");
      // if it fails we take the next file
      if(f==0) {
        fprintf(stderr,"dump: could not open %s\n", argv[argnum]);
        continue;
      }

      if(beginMode || endMode)
        printf("dumping %i bytes of %s\n", dumpLength, argv[argnum]);
      else
        printf("dumping %s\n", argv[argnum]);

      // dump the stream. If errors appeared we print that to stderr
      int read=dump(f, beginMode, endMode);
      if(read==-1) {
        fprintf(stderr,"dump: error while dumping %s\n", argv[argnum]);
        ret=1;
      }
      else
        printf("%s size: 0x%X = %i\n\n", argv[argnum], read, read);

      fclose(f);
    }
  }

  return ret;
}
